use std::{collections::HashMap, sync::Arc};

use axum::{
    extract::Query,
    routing::{get, post},
    Extension,
};
// use neo_entity::{
//     devices::{self, Entity as Device},
//     sessions::{self, Entity as Session},
//     users::{self, Entity as User},
// };

use neo_entity::prelude::*;
use sea_orm::{ActiveModelTrait, ColumnTrait, DatabaseConnection, EntityTrait, QueryFilter, Set};

use crate::{
    api::client_server::errors::{
        api_error::ApiError, authentication_error::AuthenticationError,
        registration_error::RegistrationError,
    },
    ruma_wrapper::{RumaRequest, RumaResponse},
    Config,
};

use ruma::api::client::{
    account, session,
    uiaa::{IncomingAuthData, IncomingUserIdentifier},
};

pub fn routes() -> axum::Router {
    axum::Router::new()
        .route("/r0/login", get(get_login_types).post(login))
        .route("/r0/register", post(post_register))
        .route("/r0/register/available", get(get_username_available))
}

async fn get_login_types() -> Result<RumaResponse<session::get_login_types::v3::Response>, ApiError>
{
    use session::get_login_types::v3::*;

    Ok(RumaResponse(session::get_login_types::v3::Response::new(
        vec![LoginType::Password(PasswordLoginType::new())],
    )))
}

async fn login(
    Extension(config): Extension<Arc<Config>>,
    Extension(db): Extension<DatabaseConnection>,
    RumaRequest(req): RumaRequest<session::login::v3::IncomingRequest>,
) -> Result<RumaResponse<session::login::v3::Response>, ApiError> {
    use session::login::v3::*;

    match req.login_info {
        IncomingLoginInfo::Password(incoming_password) => {
            let password = incoming_password.password;
            let user_id = if let IncomingUserIdentifier::UserIdOrLocalpart(user_id) =
                incoming_password.identifier
            {
                ruma::UserId::parse_with_server_name(user_id, &config.server_name)
                    .map_err(|_| AuthenticationError::InvalidUserId)?
            } else {
                return Err(AuthenticationError::InvalidUserId.into());
            };

            let db_user = User::find()
                .filter(users::Column::UserId.eq(user_id.as_str()))
                .one(&db)
                .await?
                .ok_or(AuthenticationError::InvalidUserId)?;
            // TODO: check password
            //db_user
            //    .password_correct(&password)
            //    .map_err(|_| AuthenticationError::Forbidden)?;

            let device = if let Some(device_id) = req.device_id {
                Device::find()
                    .filter(
                        devices::Column::DeviceId
                            .eq(device_id.as_str())
                            .and(devices::Column::UserUuid.eq(db_user.uuid)),
                    )
                    .one(&db)
                    .await?
                    .unwrap()
                //Device::find_for_user(&db, &db_user, device_id.as_str()).await?
            } else {
                let device_id = uuid::Uuid::new_v4().to_string();
                let display_name = req
                    .initial_device_display_name
                    .unwrap_or_else(|| "Generic Device".into());
                let device = devices::ActiveModel {
                    device_id: Set(device_id),
                    display_name: Set(display_name),
                    user_uuid: Set(db_user.uuid),
                    ..Default::default()
                };
                device.insert(&db).await?
            };

            let session = sessions::ActiveModel {
                device_uuid: Set(device.uuid),
                ..Default::default()
            };
            let session = session.insert(&db).await?;
            let response = Response::new(
                user_id,
                session.key,
                ruma::OwnedDeviceId::from(device.device_id),
            );

            Ok(RumaResponse(response))
        }
        _ => todo!(),
    }
}

async fn get_username_available(
    Extension(config): Extension<Arc<Config>>,
    Extension(db): Extension<DatabaseConnection>,
    Query(params): Query<HashMap<String, String>>,
) -> Result<RumaResponse<account::get_username_availability::v3::Response>, ApiError> {
    use account::get_username_availability::v3::*;
    tracing::debug!("username_available hit");

    let username = params
        .get("username")
        .ok_or(RegistrationError::MissingUserId)?
        .to_owned();
    let user_id = ruma::UserId::parse_with_server_name(username, &config.server_name)
        .map_err(|_| RegistrationError::InvalidUserId)?;

    let available = User::find()
        .filter(users::Column::UserId.eq(user_id.as_str()))
        .one(&db)
        .await?
        .is_none();

    Ok(RumaResponse(Response::new(available)))
}

async fn post_register(
    Extension(config): Extension<Arc<Config>>,
    Extension(db): Extension<DatabaseConnection>,
    RumaRequest(req): RumaRequest<account::register::v3::IncomingRequest>,
) -> Result<RumaResponse<account::register::v3::Response>, ApiError> {
    use account::register::v3::*;

    config
        .enable_registration
        .then(|| true)
        .ok_or(RegistrationError::RegistrationDisabled)?;

    match req.auth {
        Some(auth) => match auth {
            IncomingAuthData::Password(incoming_password) => {
                let password = incoming_password.password;
                let user_id = if let IncomingUserIdentifier::UserIdOrLocalpart(user_id) =
                    incoming_password.identifier
                {
                    ruma::UserId::parse_with_server_name(user_id, &config.server_name)
                        .map_err(|_| AuthenticationError::InvalidUserId)?
                } else {
                    return Err(AuthenticationError::InvalidUserId.into());
                };

                User::find()
                    .filter(users::Column::UserId.eq(user_id.as_str()))
                    .one(&db)
                    .await?
                    .ok_or(RegistrationError::UserIdTaken)?;

                let display_name = req
                    .initial_device_display_name
                    .unwrap_or_else(|| "Random Display Name".into());

                // TODO: Hash password
                let user = users::ActiveModel {
                    user_id: Set(user_id.to_string()),
                    display_name: Set(user_id.to_string()),
                    password_hash: Set(password),
                    ..Default::default()
                }
                .insert(&db)
                .await?;

                let device = devices::ActiveModel {
                    display_name: Set(display_name),
                    user_uuid: Set(user.uuid),
                    ..Default::default()
                }
                .insert(&db)
                .await?;

                let mut response = Response::new(
                    ruma::UserId::parse(&user.user_id).map_err(|e| anyhow::anyhow!(e))?,
                );
                if !req.inhibit_login {
                    let session = sessions::ActiveModel {
                        device_uuid: Set(device.uuid),
                        ..Default::default()
                    }
                    .insert(&db)
                    .await?;
                    response.access_token = Some(session.key);
                }
                if !req.inhibit_login {
                    response.device_id = Some(device.device_id.into());
                }

                Ok(RumaResponse(response))
            }
            _ => todo!(),
        },
        // For clients not following using UIAA
        None => {
            let password = req
                .password
                .ok_or("password missing")
                .map_err(|_| RegistrationError::AdditionalAuthenticationInformation)?;
            let user_id = if let Some(username) = req.username {
                ruma::UserId::parse_with_server_name(username, &config.server_name)
                    .map_err(|e| anyhow::anyhow!(e))?
            } else {
                return Err(AuthenticationError::InvalidUserId.into());
            };

            User::find()
                .filter(users::Column::UserId.eq(user_id.as_str()))
                .one(&db)
                .await?
                .ok_or(RegistrationError::UserIdTaken)?;

            let display_name = req
                .initial_device_display_name
                .unwrap_or_else(|| "Random Display Name".into());

            let user = users::ActiveModel {
                user_id: Set(user_id.to_string()),
                display_name: Set(user_id.to_string()),
                password_hash: Set(password),
                ..Default::default()
            }
            .insert(&db)
            .await?;

            let device = devices::ActiveModel {
                display_name: Set(display_name),
                user_uuid: Set(user.uuid),
                ..Default::default()
            }
            .insert(&db)
            .await?;

            let mut response =
                Response::new(ruma::UserId::parse(&user.user_id).map_err(|e| anyhow::anyhow!(e))?);
            if !req.inhibit_login {
                let session = sessions::ActiveModel {
                    device_uuid: Set(device.uuid),
                    ..Default::default()
                }
                .insert(&db)
                .await?;
                response.access_token = Some(session.key);
            }
            if !req.inhibit_login {
                response.device_id = Some(ruma::OwnedDeviceId::from(device.device_id));
            }

            Ok(RumaResponse(response))
        }
    }
}
