use std::sync::Arc;

use axum::{
    body::Body,
    handler::Handler,
    http::{Request, StatusCode},
    Extension, Router,
};
use config::Config;
use sea_orm::Database;
use tower_http::{
    cors::CorsLayer,
    trace::{DefaultMakeSpan, DefaultOnRequest, DefaultOnResponse, TraceLayer},
};
use tracing::Level;
use neo_migration::{Migrator, MigratorTrait};

mod api;
mod config;
mod responses;
mod ruma_wrapper;
mod types;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    if std::env::var("RUST_LOG").is_err() {
        std::env::set_var("RUST_LOG", "info,sqlx=off");
    }

    tracing_subscriber::fmt::init();

    let config = Arc::new(Config::default());

    let pool = Database::connect(config.db_path.clone()).await?;
    Migrator::up(&pool, None).await?;

    // TODO: set correct CORS headers
    let cors = CorsLayer::new()
        .allow_origin(tower_http::cors::Any)
        .allow_methods(tower_http::cors::Any)
        .allow_headers(tower_http::cors::Any);

    let tracing_layer = TraceLayer::new_for_http()
        .make_span_with(DefaultMakeSpan::new().level(Level::INFO))
        .on_request(DefaultOnRequest::default().level(Level::INFO))
        .on_response(DefaultOnResponse::default().level(Level::INFO));

    let client_server = Router::new()
        .merge(api::client_server::versions::routes())
        .merge(api::client_server::r0::auth::routes())
        .merge(api::client_server::r0::thirdparty::routes())
        .merge(api::client_server::r0::room::routes())
        .merge(api::client_server::r0::presence::routes())
        .merge(api::client_server::r0::push::routes())
        .merge(api::client_server::r0::filter::routes())
        .merge(api::client_server::r0::sync::routes())
        .merge(api::client_server::r0::keys::routes());

    let router = Router::new()
        .nest("/_matrix/client", client_server)
        .layer(cors)
        .layer(tracing_layer)
        .fallback(fallback.into_service())
        .layer(Extension(pool))
        .layer(Extension(config));

    let _ = axum::Server::bind(&"127.0.0.1:3000".parse().unwrap())
        .serve(router.into_make_service())
        .await?;

    Ok(())
}

async fn fallback(request: Request<Body>) -> StatusCode {
    tracing::error!("{} {}", request.method(), request.uri());
    StatusCode::INTERNAL_SERVER_ERROR
}
