use axum::{
    body::{Bytes, Full, HttpBody},
    extract::{FromRequest, Path},
    response::IntoResponse,
    BoxError,
};
use http::StatusCode;
use ruma::{
    api::{IncomingRequest, OutgoingResponse},
    exports::bytes::{BufMut, BytesMut},
    serde::CanonicalJsonValue,
};

use crate::api::client_server::errors::api_error::ApiError;

pub struct RumaRequest<R>(pub R)
where
    R: IncomingRequest;

#[axum::async_trait]
impl<R, B> FromRequest<B> for RumaRequest<R>
where
    R: IncomingRequest,
    B: HttpBody + Send,
    B::Data: Send,
    B::Error: Into<BoxError>,
{
    type Rejection = ApiError;

    async fn from_request(
        req: &mut axum::extract::RequestParts<B>,
    ) -> Result<Self, Self::Rejection> {
        let path_params = Path::<Vec<String>>::from_request(req)
            .await
            .map_err(|e| anyhow::anyhow!(e))?;
        let body = Bytes::from_request(req)
            .await
            .map_err(|e| anyhow::anyhow!(e))?;

        let json = serde_json::from_slice::<CanonicalJsonValue>(&body).ok();

        let mut buf = BytesMut::new().writer();
        serde_json::to_writer(&mut buf, &json).expect("can't fail");
        let body = buf.into_inner().freeze();

        let builder = http::Request::builder().uri(req.uri()).method(req.method());
        let request = builder.body(body).map_err(|e| anyhow::anyhow!(e))?;

        Ok(Self(
            R::try_from_http_request(request, &path_params).map_err(|e| anyhow::anyhow!(e))?,
        ))
    }
}

pub struct RumaResponse<R>(pub R)
where
    R: OutgoingResponse;

impl<R: OutgoingResponse> IntoResponse for RumaResponse<R> {
    fn into_response(self) -> axum::response::Response {
        match self.0.try_into_http_response::<BytesMut>() {
            Ok(res) => res.map(BytesMut::freeze).map(Full::new).into_response(),
            Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
        }
    }
}
