use std::net::{Ipv4Addr, Ipv6Addr, SocketAddrV4, SocketAddrV6};

#[derive(Debug, PartialEq, Eq)]
pub enum Hostname {
    IPv4(Ipv4Addr),
    IPv6(Ipv6Addr),
    Fqdn(String),
}

pub struct ServerName {
    pub hostname: Hostname,
    pub port: Option<u16>,
}

impl ServerName {
    pub fn new(server_name: &str) -> anyhow::Result<Self> {
        if let Ok(addr) = server_name.parse::<Ipv4Addr>() {
            return Ok(Self {
                hostname: Hostname::IPv4(addr),
                port: None,
            });
        };

        if let Ok(addr) = server_name.parse::<SocketAddrV4>() {
            return Ok(Self {
                hostname: Hostname::IPv4(addr.ip().to_owned()),
                port: Some(addr.port()),
            });
        };

        if server_name.ends_with(']') {
            if let Ok(addr) = server_name[1..server_name.chars().count() - 1].parse::<Ipv6Addr>() {
                return Ok(Self {
                    hostname: Hostname::IPv6(addr),
                    port: None,
                });
            };
        }

        if let Ok(addr) = server_name.parse::<SocketAddrV6>() {
            return Ok(Self {
                hostname: Hostname::IPv6(addr.ip().to_owned()),
                port: Some(addr.port()),
            });
        };

        if let Some(idx) = server_name.find(':') {
            let (hostname, port) = server_name.split_at(idx);
            Ok(Self {
                hostname: Hostname::Fqdn(hostname.to_owned()),
                port: Some(port[1..].parse()?),
            })
        } else {
            Ok(Self {
                hostname: Hostname::Fqdn(server_name.to_owned()),
                port: None,
            })
        }
    }
}

impl std::fmt::Display for ServerName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match &self.hostname {
            Hostname::IPv4(hostname) => write!(f, "{hostname}"),
            Hostname::IPv6(hostname) => write!(f, "[{hostname}]"),
            Hostname::Fqdn(hostname) => write!(f, "{hostname}"),
        }?;

        if let Some(port) = self.port {
            write!(f, ":{port}")?;
        }
        Ok(())
    }
}

mod tests {
    use super::*;

    #[test]
    fn parse_ipv4_without_port() {
        let server_name = ServerName::new("127.0.0.1").unwrap();
        assert_eq!(
            server_name.hostname,
            Hostname::IPv4("127.0.0.1".parse().unwrap())
        );
        assert_eq!(server_name.port, None);
    }

    #[test]
    fn parse_ipv4_with_port() {
        let server_name = ServerName::new("127.0.0.1:8080").unwrap();
        assert_eq!(
            server_name.hostname,
            Hostname::IPv4("127.0.0.1".parse().unwrap())
        );
        assert_eq!(server_name.port, Some(8080));
    }

    #[test]
    fn parse_ipv6_without_port() {
        let server_name = ServerName::new("[::1]").unwrap();
        assert_eq!(server_name.hostname, Hostname::IPv6("::1".parse().unwrap()));
        assert_eq!(server_name.port, None);
    }

    #[test]
    fn parse_ipv6_with_port() {
        let server_name = ServerName::new("[::1]:8080").unwrap();
        assert_eq!(server_name.hostname, Hostname::IPv6("::1".parse().unwrap()));
        assert_eq!(server_name.port, Some(8080));
    }

    #[test]
    fn parse_fqdn_without_port() {
        let server_name = ServerName::new("example.com").unwrap();
        assert_eq!(server_name.hostname, Hostname::Fqdn("example.com".into()));
        assert_eq!(server_name.port, None);
    }

    #[test]
    fn parse_fqdn_with_port() {
        let server_name = ServerName::new("example.com:8080").unwrap();
        assert_eq!(server_name.hostname, Hostname::Fqdn("example.com".into()));
        assert_eq!(server_name.port, Some(8080));
    }
}
